%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% Osm dam
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Autor: Pavel Zupa <pavel.zupa@gmail.com>
%
% Progam hleda nekonfliktni rozestaveni osmi dam metodou hill climbing
%
%
%

% Reprezentace stavu:
% s(Attacks,[Y1,Y2,...,Yn]) 
%   - Attacks - pocet vzajemnych utoku dam (hodnota objektivni funkce)
%   - Yx - cisla radku na kterych stoji jednotlive damy

:- use_module(library(lists)).
:- use_module(library(random)).

:- dynamic maxSteps/1.

size(8).

zeroList(0,[]).
zeroList(Length, [0|List]) :-
	L2 is Length-1,
	zeroList(L2,List).

% Attacks je pocet utoku v konfiguraci Queens
countAttacks(s(Attacks,Queens)) :-
	countAttacks(Queens,0,Attacks).	

countAttacks([],Attacks,Attacks).
countAttacks([Y|Tail],Accum, Attacks) :-
	numAttacks(Y,[Y|Tail],Att),
	Accum1 is Accum + Att,
	countAttacks(Tail,Accum1,Attacks).
	
% Attacks je pocet utoku, ktere jdou na policko na souradnici 0,Y z dalsich sloupcu (vice vpravo).
% 0,Y znamena, ze musim dostat pole kraloven ukrojene tak, aby ta cilova byla prvni.
numAttacks(_,[],0).
numAttacks(Y,[_|Queens],Attacks) :-
	horizAttacks(Y,Queens,0,Attacks1),
	diag1Attacks(Y,1,Queens,Attacks1,Attacks2),
	diag2Attacks(Y,1,Queens,Attacks2,Attacks),!.

horizAttacks(_,[],A,A).
horizAttacks(Y,[Q|Queens],Accum,Attacks) :-
	Y is Q,
	NewAccum is Accum+1,
	horizAttacks(Y,Queens,NewAccum,Attacks);
	horizAttacks(Y,Queens,Accum,Attacks).

diag1Attacks(_,_,[],A,A).
diag1Attacks(Y,XDist,[Q|Queens],Accum,Attacks) :-
	Y is Q-XDist,
	NewAccum is Accum+1,
	XDist2 is XDist+1,
	diag1Attacks(Y,XDist2,Queens,NewAccum,Attacks);
	XDist2 is XDist+1,
	diag1Attacks(Y,XDist2,Queens,Accum,Attacks).
diag2Attacks(_,_,[],A,A).
diag2Attacks(Y,XDist,[Q|Queens],Accum,Attacks) :-
	Q is Y-XDist,
	NewAccum is Accum+1,
	XDist2 is XDist+1,
	diag2Attacks(Y,XDist2,Queens,NewAccum,Attacks);
	XDist2 is XDist+1,
	diag2Attacks(Y,XDist2,Queens,Accum,Attacks).

printState(s(Attacks,Queens)) :-
	size(X),
	printState(Queens,0,X),
	printLine(Queens),write('+ Attacks: '),write(Attacks),nl,!.

printState(_,X,X).
printState(Queens,A,Size) :-
	printLine(Queens),write('+'),nl,
	printRow(Queens,A),write('|'),nl,
	A2 is A+1,
	printState(Queens,A2,Size).

printRow([],_).
printRow([A|Queens],A) :-
	write('| O '),
	printRow(Queens,A).
printRow([_|Queens],A) :-
	write('|   '),
	printRow(Queens,A).

printLine([]).
printLine([_|Queens]) :-
	write('+---'),
	printLine(Queens).

%hlavni algoritmus hill climbing
hillClimbing(Queens,MaxSteps, Steps, Solution) :-
	State = s(_,Queens),
	countAttacks(State),
	climb(State, MaxSteps, 0, Steps, Solution).

% nasli jsme reseni
climb(s(0,Queens),_,Steps,Steps,s(0,Queens)).
% prekrocili jsme maxmalni pocet kroku vypoctu
climb(S,Steps,Steps,Steps,S).

climb(s(Attacks,Queens),MaxSteps,Step,Steps,Solution) :-
	bestMove(Queens,s(NewAttacks,NewQueens)),
	NewAttacks =< Attacks,
	Step1 is Step+1,
	%write(Step1),write(': '),write(s(NewAttacks,NewQueens)),nl,
	climb(s(NewAttacks,NewQueens),MaxSteps,Step1,Steps,Solution),!;
	climb(s(Attacks,Queens),Step,Step,Steps,Solution).


% BestMove je nejlepsi mozny tah z pozice Queens
% v pripade vice stejne dobrych nejlepsich tahu je vysledkem nahodny z nich
bestMove(Queens,BestMove) :-
	moves(Queens,Moves),
	bestSet(Moves,BestMoves),
	length(BestMoves,L),
	random(0,L,N),
	nth0(N,BestMoves,BestMove).
	
% Moves je seznam vsech pozic dosazitelnych pohybem jedne damy z pozice Queens
moves(Queens,Moves) :-
	size(S),
	moves(S,Queens,[],Moves).
	
moves(0,_,Moves,Moves).
moves(N,Queens,Accum,Moves) :-
	N1 is N-1,
	nth0(N1,Queens,Y,NewQueens),
	incList(Y,Nums),
	rowMoves(Nums, N1, NewQueens, [], NewNewQueens),
	append(Accum,NewNewQueens,NewAccum),!,
	moves(N1,Queens,NewAccum, Moves).

rowMoves([],_,_,Q,Q).
rowMoves([Y|YTail], N, Queens, Accum, Moves) :-
	nth0(N,NewQueens,Y,Queens),
	countAttacks(s(A,NewQueens)),
	rowMoves(YTail, N, Queens, [s(A,NewQueens)|Accum], Moves).

% BestMoves je seznam nejlepsich tahu z Moves 
% (tzn. jedna nebo vice pozic, ktere maji minimalni pocet utoku)
bestSet(Moves,BestMoves) :-
	bestSet(Moves,1000,[],BestMoves).

bestSet([],_,BestMoves,BestMoves).
bestSet([s(A,Q)|Moves],Min,Accum,BestMoves) :-
	A < Min,
	bestSet(Moves,A,[s(A,Q)],BestMoves);
%bestSet([s(A,Q)|Moves],Min,Accum,BestMoves) :-
	A == Min,
	bestSet(Moves,A,[s(A,Q)|Accum],BestMoves);
%bestSet([s(A,_)|Moves],Min,Accum,BestMoves) :-
	A > Min,
	bestSet(Moves,Min,Accum,BestMoves).
	
% List je seznam vzestupnych cisel 0 az size(), kde chybi cislo Exclude
incList(Exclude,List) :-
	size(S),
	incList(0,S,Exclude, List).
incList(S,S,_,[]).
incList(X,S,X,List) :-
	NewX is X+1,
	incList(NewX,S,X,List).
incList(X,S,A,[X|List]) :-
	X < S,
	NewX is X+1,
	incList(NewX,S,A,List).

% S je nahodny stav velikosti Size
randomState(Size,S) :-
	createNums(Size,0,[],S).
% S je nahodny stav velikosti Size s Conflicts poctem vzajemne utocicich paru dam
randomState(Size,Conflicts,S) :-
	createNums(Size,0,[],S),
	countAttacks(s(C,S)),
	C == Conflicts;
	randomState(Size,Conflicts,S),!.

% vytvori nahodny stav
createNums(0,_,Nums,Nums) :-!.
createNums(Cnt,Size,Nums,New) :-
	N is Cnt-1,
	Size2 is Size+1,
	random(0,Size2,I),
	nth0(I,Neww,N,Nums),
	createNums(N,Size2,Neww,New).

stats(Conflicts, Count, MaxSteps, SuccCnt, SuccAvg, SuccMin, SuccMax):-
	solveMore(Conflicts,Count,MaxSteps,[],Sols),
	succCount(Sols,SuccCnt),
	succSum(Sols,SuccSum),
	SuccAvg is SuccSum / SuccCnt,
	succMin(Sols,SuccMin),
	succMax(Sols,SuccMax),!.
	
solveMore(_,0,_,Sols,Sols).
solveMore(Conflicts, Count, MaxSteps, Accum, Sols) :-
	size(Size),
	randomState(Size,Conflicts,State),
	hillClimbing(State,MaxSteps, Steps, s(Sol,_)),
	Count1 is Count-1,
	solveMore(Conflicts, Count1,MaxSteps,[t(Steps,Sol)|Accum],Sols).
	
succCount(Sols,Cnt) :-
	succCount(Sols,0,Cnt).
succCount([],A,A).
succCount([t(_,0)|Tail],Accum,Count) :-
	NewAccum is Accum+1,
	succCount(Tail,NewAccum,Count).

succCount([t(_,X)|Tail],Accum,Count) :-
	X > 0,
	succCount(Tail,Accum,Count).

succSum(Sols,Cnt) :-
	succSum(Sols,0,Cnt).
succSum([],A,A).
succSum([t(A,0)|Tail],Accum,Count) :-
	NewAccum is Accum+A,
	succSum(Tail,NewAccum,Count).
succSum([t(_,X)|Tail],Accum,Count) :-
	X > 0,
	succSum(Tail,Accum,Count).

succMin(Sols,Min) :-
	succMin(Sols,1000,Min).
succMin([],A,A).
succMin([t(A,0)|Tail],Accum,Min) :-
	A < Accum,
	succMin(Tail,A,Min);
	succMin(Tail,Accum,Min).
succMin([t(_,X)|Tail],Accum,Min) :-
	X > 0,
	succMin(Tail,Accum,Min).

succMax(Sols,Max) :-
	succMax(Sols,0,Max).
succMax([],A,A).
succMax([t(A,0)|Tail],Accum,Max) :-
	A > Accum,
	succMax(Tail,A,Max);
	succMax(Tail,Accum,Max).
succMax([t(_,X)|Tail],Accum,Max) :-
	X > 0,
	succMax(Tail,Accum,Max).

testAtoB(B,B).

testAtoB(A,B) :-
	stats(A, 100, 100, SuccCnt, SuccAvg, SuccMin, SuccMax),
	write('Attacks: '),write(A),write('  Succ: '),write(SuccCnt),write('  Avg: '),write(SuccAvg),write('  Min: '),write(SuccMin),write('  Max: '),write(SuccMax),nl,
	A1 is A+1,
	testAtoB(A1,B).
	
	
